/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#include "h2d.h"
#include "h2d_struct.h"
#include "h2d_xlat.h"
#include "h2d_echo.h"
#include "h2d_transform.h"
#ifdef __linux__
#include "h2d_linux.h"
#endif

#include <sstream>

using namespace std;
using namespace boost::spirit;

using namespace h2d;

//////////////////////////////////////////////////////////////////////////////////////////
//
// <c>.h to .d module thang
// Takes the output of a C preprocessor (linux only tested atm) and messes
// around with it until it can be used as a D module.
// Currently written using the C++ boost::spirit library (which is fairly 
// cool btw - and I'm also fairly sure I could use it alot better than I am,
// but its a first try)
// This a 4 pass process right now because I haven't figured out how to
// apply two transforms to the same piece of text at the same time without
// too much messing about. eg. transforming the types in a 'typedef struct'
// while also re-writing the struct definition.
// First pass reads directly from the specified input file (can't be std::cin
// because spirit can't parse a std::istream) and writes to a string
// stream which is then read by the second pass. The last pass writes
// directly to std::cout.
// More passes may become necessary as more features are added.
// Implementation is via boost::spirit::grammar classes which allow you to
// mess with the input iterators without having to rewrite stuff when you
// decide to change things around - bit more code up front, less messing
// later on.
// NOTE: Yes this code could be a lot less 'wordy' - I prefer to be able to
// understand whats going though. You'll have to bare with me.
//



int main(int argc, const char** argv) {

  if (argc < 2) {
    cout << "Usage: " << argv[0] << " [options] <filename>" << endl;
    return 1;
  }

  iterator_t file(argv[argc - 1]);
  if (!file) {
    cout << "Unable to open file: " << argv[argc - 1] << endl;
    return 1;
  }

  ostringstream output;

  // Redirect std::cout to output (save original buffer)
  std::streambuf* coutSave = cout.rdbuf();
  cout.rdbuf(output.rdbuf());

  // First pass reads from the input file and outputs to 
  // the std::ostingstream held in 'output'
  // 1st Pass: xlats and echo
  
  cerr << "Pass 1" << endl;
  
  XlatGrammar      xlat;
  EchoGrammar      echo;
  parse(file, file.make_end(), *(xlat | echo));

  std::string pass1 = output.str();
  output.str(string());

  // Second pass reads from the string produced by the first
  // pass. Messing with structs & unions
  cerr << "Pass 2" << endl;

  StructGrammar    structs;
  parse(pass1.c_str(), *(structs | echo));


  // Third pass...
  cerr << "Pass 3" << endl;
  std::string pass2 = output.str();
  output.str(string());

  TransformGrammar transforms;
  parse(pass2.c_str(), *(transforms | echo));

  // Nods towards the poor lost souls on Windows :)
#ifdef __linux__
  // Fourth! pass... This could have been done in the third
  // pass but I wanted to keep the platform/compiler stuff 
  // separate
  //
  cerr << "Pass 4" << endl;

  std::string pass3 = output.str();
  output.str(string());

  LinuxGrammar linuxisms;
  parse(pass3.c_str(), *(linuxisms | echo));
#endif

  // Dump the output from the last pass to cout
  cout.rdbuf(coutSave);
  cout << "extern (C) {" << endl;
  cout << output.str();
  cout << "}" << endl;

  return 0;
}


