/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#ifndef H2D_ECHO_H
#define H2D_ECHO_H

#include <iostream>

// Much of the header files being processed need only be spat back out again
// as they were found.
//

namespace h2d {

  struct EchoGrammar : public boost::spirit::grammar<EchoGrammar> {

    template<typename T>
    struct definition {

      typedef boost::spirit::rule<T> rule_t;

      definition(EchoGrammar const& self) {
        // The echo grammar is always the last parse match possible. Its reason
        // for existence is purely to pickup what the rest of the parse rules don't
        // want to handle and to spit the text back out into the next pass.
        // The way these rules are formulated causes the parser to drop back a level
        // whenever the character types being parsed change. So when a piece of whitespace
        // finishes and some text begins the parser will drop out of this match and see
        // if the text is relevant to another parse rule - that way we don't get runaway
        // echo matches which just dump the entire parse text to std::cout.
        //
        punctuation_ = chset_p("!\"`'%^&*()-+={}[]:;~?<>,./\\|");
        text_ = print_p - (space_p | punctuation_);
        rule_ = (punctuation_ | +text_ | +space_p)[boost::bind(&definition<T>::echo, _1, _2)];
      }

      rule_t const& start() const { return rule_; }

    private:
      static void echo(typename T::iterator_t begin, typename T::iterator_t end) {
        std::cout << std::string(begin, end);
      }

      rule_t punctuation_;
      rule_t text_;
      rule_t rule_;
    };
  };
}

#endif
