/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#ifndef H2D_LINUX_H
#define H2D_LINUX_H

#include "h2d.h"
#include "h2d_d.h"

//
// h2d_linux.h: All the bits and bobs which are linux specific and
// need special case fixups. These are fixups that I recon will be
// needed in more than one header. ie do it here instead of in a patch
//

namespace h2d {

  struct LinuxGrammar : public boost::spirit::grammar<LinuxGrammar> {
    template<typename T>
    struct definition {

      typedef boost::spirit::rule<T> rule_t;

      definition(LinuxGrammar const& self) {
        
        // Delete function predeclarations which are followed by inline
        // definitions.
        // I don't know how feasible this will be long term but I suspect
        // that there will only be a few specific deletions like this...
        parens_ = ch_p('(') >> (parens_ | *(anychar_p - chset_p("()"))) >> ch_p(')');
        linux_decls_ =
          str_p("gnu_dev_major") |
          str_p("gnu_dev_minor") |
          str_p("gnu_dev_makedev");
        erase_decls_ = 
          !(str_p("extern") >> +space_p) >> 
          d_types_ >> +space_p >>
          linux_decls_ >> *space_p >>
          parens_ >> *space_p >>
          ch_p(';');

        rule_ = erase_decls_;
      }

      rule_t const& start() const { return rule_; }
    private:
      rule_t       rule_;
      rule_t       parens_;
      rule_t       linux_decls_;
      rule_t       erase_decls_;
      
      DTypesGrammar d_types_;
    };
  };
}


#endif
