/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#ifndef H2D_XLAT_H
#define H2D_XLAT_H

#include "h2d.h"
#include "h2d_d.h"

#define BOOST_SPIRIT_NO_REGEX_LIB
#include <boost/spirit/utility/regex.hpp>

#include <iostream>

//////////////////////////////////////////////////////////////////////////////////////////
//
// Simple transliterations from one string of word(s) to another - which
// also includes string deletions (xlat to null)
//

namespace h2d {

  inline void xlat(const char* out) {
    std::cout << out;
  }
  
  struct XlatGrammar : public boost::spirit::grammar<XlatGrammar> {
    template<typename T>
    struct definition {

      typedef typename T::iterator_t iterator_t;
      
      definition(XlatGrammar const& self) {

        // function -> Function
        // class    -> Class
        // new      -> New
        // These are D keywords - sometimes they are used for member/parameter names in C
        // Not sure whether uppercasing the first letter is a brilliant idea but it keeps the
        // name as close to that intended as possible
        delimiters_ = space_p | chset_p(";,()[]{}-+=*&^%!'~:<>./?\"");
        xlat_keywords_ = 
          delimiters_[boost::bind(&definition<T>::keywordBE, this, _1, _2)] 
            >> 
          ((str_p("function")[boost::bind(&definition<T>::keywordS, this, std::string("Function"))])
           | 
           (str_p("class")[boost::bind(&definition<T>::keywordS, this, std::string("Class"))])
           |
           (str_p("version")[boost::bind(&definition<T>::keywordS, this, std::string("Version"))])
           |
           (str_p("new")[boost::bind(&definition<T>::keywordS, this, std::string("New"))]))
            >> 
          delimiters_[boost::bind(&definition<T>::keywordE, this, _1, _2)];
        
        // (void) -> ()
        xlat_void_ = (ch_p('(') >> *space_p >> str_p("void") >> *space_p >> ch_p(')'))[boost::bind(&xlat, "()")];

        // typedef -> alias
        xlat_alias_ = str_p("typedef")[boost::bind(&xlat, "alias")];
        
        // long double -> real
        xlat_real_ = (str_p("long") >> +space_p >> str_p("double"))
          [boost::bind(xlat, "real")];
        
        // unsigned long long int -> ulong
        xlat_ulong_a_ = (str_p("unsigned") >> +space_p >> str_p("long") >> +space_p >> str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "ulong")];
        // unsigned long long -> ulong
        xlat_ulong_b_ = (str_p("unsigned") >> +space_p >> str_p("long") >> +space_p >> str_p("long"))
          [boost::bind(xlat, "ulong")];
        // signed long long int -> long
        xlat_long_a_ = (str_p("signed") >> +space_p >> str_p("long") >> +space_p >> str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "long")];
        // signed long long -> long
        xlat_long_b_ = (str_p("signed") >> +space_p >> str_p("long") >> +space_p >> str_p("long"))
          [boost::bind(xlat, "long")];
        // long long int -> long
        xlat_long_c_ = (str_p("long") >> +space_p >> str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "long")];
        // long long -> long
        xlat_long_d_ = (str_p("long") >> +space_p >> str_p("long"))
          [boost::bind(xlat, "long")];
        
        // unsigned long int -> uint
        xlat_uint_a_ = (str_p("unsigned") >> +space_p >> str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "uint")];
        // unsigned long -> uint
        xlat_uint_b_ = (str_p("unsigned") >> +space_p >> str_p("long"))
          [boost::bind(xlat, "uint")];
        // unsigned int -> uint
        xlat_uint_c_ = (str_p("unsigned") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "uint")];
        // signed long int -> int
        xlat_int_a_ = (str_p("signed") >> +space_p >> str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "int")];
        // signed long -> int
        xlat_int_b_ = (str_p("signed") >> +space_p >> str_p("long"))
          [boost::bind(xlat, "int")];
        // signed int -> int
        xlat_int_c_ = (str_p("signed") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "int")];
        // long int -> int
        xlat_int_d_ = (str_p("long") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "int")];
        // long -> int
        xlat_int_e_ = str_p("long")[boost::bind(xlat, "int")];
        // bool -> int
        xlat_int_f_ = str_p("bool")[boost::bind(xlat, "int")];
        
        
        // unsigned short int -> ushort
        xlat_ushort_a_ = (str_p("unsigned") >> +space_p >> str_p("short") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "ushort")];
        // unsigned short -> ushort
        xlat_ushort_b_ = (str_p("unsigned") >> +space_p >> str_p("short"))
          [boost::bind(xlat, "ushort")];
        // signed short int -> short
        xlat_short_a_ = (str_p("signed") >> +space_p >> str_p("short") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "short")];
        // signed short -> short
        xlat_short_b_ = (str_p("signed") >> +space_p >> str_p("short"))
          [boost::bind(xlat, "short")];
        // short int -> short
        xlat_short_c_ = (str_p("short") >> +space_p >> str_p("int"))
          [boost::bind(xlat, "short")];
        
        // unsigned char -> ubyte
        xlat_ubyte_a_ = (str_p("unsigned") >> +space_p >> str_p("char"))
          [boost::bind(xlat, "ubyte")];
        // signed char -> byte
        xlat_byte_a_ = (str_p("signed") >> +space_p >> str_p("char"))
          [boost::bind(xlat, "byte")];
        // unsigned -> uint (needs a little more insurance on the end)
        xlat_unsigned_ = (str_p("unsigned") >> eps_p(~chset_p("a-zA-Z0-9_")))
          [boost::bind(xlat, "uint")];
        

        // Other deletions...
        xlat_comment_ = regex_p("^[ \t]*#") >> *print_p >> eol_p;
        xlat_parendel_ = confix_p('(', (*(anychar_p - chset_p("()"))) >> *xlat_parendel_, ')');

        // Remove any instance of the struct keyword that is not followed
        // by a definition block '{...}' - with an optional tag
        // 'struct <name>;' is also allowed
        //        xlat_struct_ = 
        //          (((str_p("struct") | str_p("union")) >> +space_p >> 
        //            *(alnum_p | '_') >> *space_p >> chset_p("{;"))[boost::bind(&definition<T>::echo, this, _1, _2)]) 
        //          |
        //          ((str_p("struct") | str_p("union")) >> +space_p);

        xlat_deletions_ = 
          xlat_comment_ | 
          //          xlat_struct_ |
          str_p("__inline") |
          str_p("__extension__") | 
          str_p("__restrict") |
          str_p("__const") |
          (str_p("__attribute__") >> *space_p >> xlat_parendel_) |
          (str_p("__attribute") >> *space_p >> xlat_parendel_);

        // Alternative xlats - the sequence here is important. 
        // They are matched first to last against the input - it is not longest or 
        // shortest match
        //
        xlats_ = xlat_deletions_ | xlat_keywords_ |
          xlat_void_ | xlat_alias_ | xlat_real_ | xlat_ulong_a_ | xlat_ulong_b_ | 
          xlat_long_a_ | xlat_long_b_ | xlat_long_c_ | xlat_long_d_ | 
          xlat_uint_a_ | xlat_uint_b_ | xlat_uint_c_ | 
          xlat_int_a_ | xlat_int_b_ | xlat_int_c_ | xlat_int_d_ | xlat_int_e_ | xlat_int_f_ | 
          xlat_ushort_a_ | xlat_ushort_b_ | xlat_short_a_ | xlat_short_b_ | xlat_short_c_ |
          xlat_ubyte_a_ | xlat_byte_a_ | xlat_unsigned_;
      }
      
      typedef boost::spirit::rule<T> rule_t;

      rule_t const& start() const { return xlats_; }
      
    private:
      // (void) -> ()
      rule_t xlat_void_;

      // typedef -> alias
      rule_t xlat_alias_;
      
      // long double -> real
      rule_t xlat_real_;
      
      // unsigned long long int -> ulong
      rule_t xlat_ulong_a_;
      // unsigned long long -> ulong
      rule_t xlat_ulong_b_;
      // signed long long int -> long
      rule_t xlat_long_a_;
      // signed long long -> long
      rule_t xlat_long_b_;
      // long long int -> long
      rule_t xlat_long_c_;
      // long long -> long
      rule_t xlat_long_d_;
      
      // unsigned long int -> uint
      rule_t xlat_uint_a_;
      // unsigned long -> uint
      rule_t xlat_uint_b_;
      // unsigned int -> uint
      rule_t xlat_uint_c_;
      // signed long int -> int
      rule_t xlat_int_a_;
      // signed long -> int
      rule_t xlat_int_b_;
      // signed int -> int
      rule_t xlat_int_c_;
      // long int -> int
      rule_t xlat_int_d_;
      // long -> int
      rule_t xlat_int_e_;
      // bool -> int
      rule_t xlat_int_f_;


      // unsigned short int -> ushort
      rule_t xlat_ushort_a_;
      // unsigned short -> ushort
      rule_t xlat_ushort_b_;
      // signed short int -> short
      rule_t xlat_short_a_;
      // signed short -> short
      rule_t xlat_short_b_;
      // short int -> short
      rule_t xlat_short_c_;
      
      // unsigned char -> ubyte
      rule_t xlat_ubyte_a_;
      // signed char -> byte
      rule_t xlat_byte_a_;
      // unsigned -> uint
      rule_t xlat_unsigned_;

      // Deletion rules
      rule_t xlat_comment_;
      rule_t xlat_deletions_;
      rule_t xlat_parendel_;
      rule_t xlat_struct_;

      rule_t xlats_;


      rule_t xlat_keywords_;
      rule_t delimiters_;
      std::string keyword_;
      
      void keywordBE(iterator_t begin, iterator_t end) {
        keyword_ = std::string(begin, end);
      }
      void keywordS(std::string const& str) {
        keyword_ += str;
      }
      void keywordE(iterator_t begin, iterator_t end) {
        std::cout << keyword_ << std::string(begin, end);
      }
      
      void echo(iterator_t begin, iterator_t end) const {
        std::cout << std::string(begin, end);
      }
    };
  };
}

#endif
